/*
 ***** BEGIN LICENSE BLOCK *****
 * This file is part of ExQuilla by Mesquilla.
 *
 * Copyright 2016 R. Kent James
 *
 * All Rights Reserved
 *
 * ***** END LICENSE BLOCK *****
 */

// provides sending of an ews message

const { classes: Cc, Constructor: CC, interfaces: Ci, utils: Cu, Exception: CE, results: Cr, } = Components;
var { XPCOMUtils } = ChromeUtils.import("resource://gre/modules/XPCOMUtils.jsm");
var { ComponentUtils } = ChromeUtils.import("resource://gre/modules/ComponentUtils.jsm");
ChromeUtils.defineModuleGetter(this, "MailServices",
  "resource:///modules/MailServices.jsm");
ChromeUtils.defineModuleGetter(this, "MessageSend",
  "resource:///modules/MessageSend.jsm");
ChromeUtils.defineModuleGetter(this, "Services",
  "resource://gre/modules/Services.jsm");

ChromeUtils.defineModuleGetter(this, "PropertyList",
                               "resource://exquilla/modules/PropertyList.jsm");
ChromeUtils.defineModuleGetter(this, "PromiseUtils",
                               "resource://exquilla/modules/PromiseUtils.jsm");
ChromeUtils.defineModuleGetter(this, "Utils",
  "resource://exquilla/modules/ewsUtils.jsm");

var _log = null;
XPCOMUtils.defineLazyGetter(this, "log", () => {
  if (!_log) _log = Utils.configureLogging("send");
  return _log;
});

// globals

var SMTP_DELIV_MAIL = "smtpDeliveringMail";

// Main class.
var global = this;
class EwsSend extends MessageSend {
  constructor() {
    if (typeof (safeGetJS) == "undefined")
      Utils.importLocally(global);

    super();

    // Thunderbird 115.2 now defines this on the object instead of the prototype
    delete this.QueryInterface;

    if (this._deliverMode == Ci.nsIMsgCompDeliverMode.Background)
    {
      log.info("Send in background not supported, just sending now");
      this._deliverMode = Ci.nsIMsgCompDeliverMode.Now;
    }

    // local instance variables
    this.mailbox = null;
    this.ewsCompose = null; // This is set by ewsCompose when creating the ewsSend object.
  }

  async _deliverAsMail()
  {
    log.config("EwsSend _deliverAsMail() deliveryMode " + this._deliverMode);
    let sendSucceeded = false;
    let sendToMailboxResult = Cr.NS_ERROR_FAILURE;

    try {
      // The item to save or send is on the current mailbox, so create it.
      let compFields = this._compFields;
      // create empty message as a property list
      let properties = oPL({ ItemClass: 'IPM.Note' });

      // append set properties
      properties.appendString('Subject', compFields.subject);
      if (compFields.references.length)
        properties.appendString('References', compFields.references);

      addAddresses(compFields.to, "ToRecipients", properties);
      addAddresses(compFields.cc, "CcRecipients", properties);
      addAddresses(compFields.bcc, "BccRecipients", properties);
      addAddresses(compFields.replyTo, "ReplyTo", properties);

      this.sendReport.currentProcess = Ci.nsIMsgSendReport.process_SMTP;
      // switch message to Delivering mail
      this.updateStatus(SMTP_DELIV_MAIL);
      // Add to address collector
      let recipients = [
        this._compFields.to,
        this._compFields.cc,
        this._compFields.bcc,
      ].filter(Boolean);
      this._collectAddressesToAddressBook(recipients);
      this.notifyListenerOnStartSending(null, null);
      let sendMessageListener = new PromiseUtils.MachineListener();
      this.mailbox.sendMessageFromFile(this._messageFile, this.mailbox.getNativeFolder("outbox"), properties, sendMessageListener);
      let sendMessageMOS = await sendMessageListener.promise; // Machine On Stop MOS
      sendToMailboxResult = sendMessageMOS.status;
      if (sendMessageMOS.status == Cr.NS_OK)
        sendSucceeded = true;
    }
    catch (e) {log.error(e);}
    finally {
      if (!sendSucceeded && sendToMailboxResult == Cr.NS_OK)
        sendToMailboxResult = Cr.NS_ERROR_FAILURE;
      log.config("Finished sending item with status " + sendToMailboxResult);
      if (sendToMailboxResult != Cr.NS_OK)
      {
        this.fail(sendToMailboxResult, this._composeBundle.GetStringFromName("sendFailed"));
      }
      // should we be doing this now, or waiting until completion?
      this.notifyListenerOnStopSending(null, sendToMailboxResult, null, null);
    }

    if (sendToMailboxResult == Cr.NS_OK)
    {
      this._doFcc();
    }
  }

  // local functions
  updateStatus(aStatusId) // see nsSmtpProtocol::UpdateStatus
  {
    let statusPosted = false;
    try {
      let progress = this._sendProgress;
      if (progress instanceof Ci.nsIMsgProgress)
      {
        let composeBundle = this._composeBundle;
        let status = (typeof aStatusId == 'number') ? composeBundle.GetStringFromID(aStatusId)
                                                    : composeBundle.GetStringFromName(aStatusId);

        progress.onStatusChange(null, null, Cr.NS_OK, status);
        statusPosted = true;
      }
      else
        log.debug("Missing MessageSend._sendProgress");
    } 
    catch (e) {log.warn("updateStatus error: " + e);}
    if (!statusPosted)
      log.warn("Did not post status for statusId " + aStatusId);
  }
}

// helper functions

// add a skink address list to a SOAP property list
function addAddresses(aList, aName, aProperties)
{
  if (!aList.length)
    return;

  let addresses = MailServices.headerParser.parseEncodedHeader(aList);

  let toPL = oPL({});
  for (let address of addresses)
  {
    let mailbox = oPL( {Name: (address.name || ""),
                        EmailAddress: address.email
                       } );
    toPL.appendPropertyList('Mailbox', mailbox);
  }
  if (addresses.length)
    aProperties.appendPropertyList(aName, toPL);
}

EwsSend.prototype.getInterface = EwsSend.prototype.QueryInterface = ChromeUtils.generateQI(["nsIMsgSend", "nsIInterfaceRequestor"]);
EwsSend.prototype.contactID = "@mesquilla.com/ewssend;1";
EwsSend.prototype.classID = Components.ID("{DBF0D7B6-C52A-4323-8B22-E4F892B4F3EE}");

var EXPORTED_SYMBOLS = ["EwsSend"];
